// -*- Mode:C++ -*-
//Header:
//File: Param.h
//Title: Parameter Struct for Chagamma
//Author: NODA, Itsuki
//Date: 2002/04/09
//

//ModifyHistory:
// 2001/04/20: Start to create this file
// 2002/04/09: divided from chagamma.cc
//EndModifyHistory:

#ifndef _Bun_PlayPhase_h_
#define _Bun_PlayPhase_h_
////////////////////////////////////////////////////////////////////////

#include "itk.h"
#include "SimplePlayer.h"

namespace Soccer {

//======================================================================
// class PlayPhase

   class PlayPhase : public WithDescriber {

      public:
	 enum Type {
	    Nop,
	    Dango,
	    FixedPosition,
	    DeployToHome,
	    PP_Max,
	 } ;
	 static const char * tname(Type t) {
	    switch(t) {
	       case Nop :			return "nop" ;
	       case Dango : 		return "dango" ;
	       case FixedPosition : 	return "fixed_pos" ;
	       case DeployToHome : 	return "deploy_to_home" ;
	       default :			return "(none)" ;
	    }
	 } ;

      public:
	 CycleStep time ;
	 Type type ;
	 Flt _farg1 ;  Flt _farg2 ;  Flt _farg3 ;
	 Int _iarg1 ;  Int _iarg2 ;  Int _iarg3 ;

	 //--------------------------------------------------
	 // access
      public:
	 ITK_DEF_ACCESS(Flt & farg1(), { return _farg1 ; }); 
	 ITK_DEF_ACCESS(Flt & farg2(), { return _farg2 ; }); 
	 ITK_DEF_ACCESS(Flt & farg3(), { return _farg3 ; }); 
	 ITK_DEF_ACCESS(Int & iarg1(), { return _iarg1 ; }); 
	 ITK_DEF_ACCESS(Int & iarg2(), { return _iarg2 ; }); 
	 ITK_DEF_ACCESS(Int & iarg3(), { return _iarg3 ; }); 

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#PlayPhase[" << tname(type) << "]" ;

	    if(detailp) {
	       ostr << endl ;
	       ostr << "\t" "farg1=" << farg1() << endl ;
	       ostr << "\t" "farg2=" << farg2() << endl ;
	       ostr << "\t" "farg3=" << farg3() << endl ;

	       ostr << "\t" "iarg1=" << iarg1() << endl ;
	       ostr << "\t" "iarg2=" << iarg2() << endl ;
	       ostr << "\t" "iarg3=" << iarg3() << endl ;
	    } 
	 } ;
   } ;

   class PlayPhaseStack : public StackT<PlayPhase> { 
      public:
	 MutexCondVar mtxcv ;
      
      public:
	 virtual ~PlayPhaseStack() {} ;

      public:
	 void lock() { return mtxcv.lock() ; } ;
	 void unlock() { return mtxcv.unlock() ; } ;
	 void signal() { return mtxcv.signal() ; } ;
	 void broadcast() { return mtxcv.broadcast() ; } ;
	 Int wait() { return mtxcv.wait() ; } ;
	 Int lockWait() { return mtxcv.lockWait() ; } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif
