// -*- Mode:C++ -*-
//Header:
//File: LogTrajPlayerT.h
//Title: Player with logging trajectry
//Author: NODA, Itsuki
//Date: 2001/03/23
//

//ModifyHistory:
// 2001/03/23: Start to create this file
//EndModifyHistory:


#ifndef _SPlayer_LogTrajPlayerT_h_
#define _SPlayer_LogTrajPlayerT_h_
////////////////////////////////////////////////////////////////////////

#include "SimplePlayer.h"
#include "OnlineCoach.h"
#include "OfflineCoach.h"

namespace Soccer {

   using namespace Itk ;

   //======================================================================
   // LogTrajPlayerT class

   /*--------------------*/
   /**
    *
    */

   template <class MotherPlayer>
   class LogTrajPlayerT : public MotherPlayer {
      public:
	 typedef LogTrajPlayerT<MotherPlayer> LogTrajPlayer ;

	 //--------------------------------------------------
	 // member
      public:
	 OnlineCoach coach ;
	 OfflineCoach control ;
	 
	 ostream * log ;
	 Bool isLogFile ;
	 Mutex mxLog ;

	 Thread * thrdLogCoachSensor ;
	 Thread * thrdLogPlayerSensor ;

	 //--------------------------------------------------
	 // destructor
      public:
	 ~LogTrajPlayerT<MotherPlayer>() {} ;
	 
	 //--------------------------------------------------
	 // init
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool init(UInt sleepMSec = 1000) {
	    isLogFile = False ;

	    control.init() ;
	    control.forkSensor() ;

	    MotherPlayer::init() ;
	    forkSensor() ;
	    
	    sleep(0,sleepMSec) ;
	    
	    coach.init() ;
	    coach.forkSensor() ;
	    
	    return True ;
	 } ;

	 //--------------------------------------------------
	 // logopen
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool openLog(const char * logfile) {
	    lockLog() ; {
	       log = new ofstream(logfile) ;
	       isLogFile = True ;
	    } ; unlockLog() ;
	    return True ;
	 } ;
	 
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool closeLog(Bool diep = False, int exitstate = 0) {
	    lockLog() ; {
	       if(isLogFile) 
		  static_cast<ofstream*>(log)->close() ;
	       if(diep) 
		  exit(exitstate) ;
	    } ; unlockLog() ;
	    return True ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void lockLog() { mxLog.lock() ; } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void unlockLog() { mxLog.unlock() ; } ;	 

	 //--------------------------------------------------
	 // send command and logging
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool turn(const Flt & power) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool dash(const Flt & power) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool kick(const Flt & power, const Flt & dir) ;

	 //--------------------------------------------------
	 // logging coach and player sensor
      public:
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool forkLogCoachSensor() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 static void * runLogCoachSensor(void * p) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void * logCoachSensor() ;
	 
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool forkLogPlayerSensor() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 static void * runLogPlayerSensor(void * p) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void * logPlayerSensor() ;

	 //--------------------------------------------------
	 // access coach's info
      public:
	 ITK_DEF_ACCESS(CycleStep & c_time(), {
	    return coach.time() ;
	 }) ;

	 ITK_DEF_ACCESS(OnlineCoach::SeeInfo::Object & c_myinfo(), {
	    return coach.sinfo.teammate(unum-1) ;
	 }) ;

	 ITK_DEF_ACCESS(PosDir & c_myposdir(), {
	    return c_myinfo().posdir() ;
	 }) ;

	 ITK_DEF_ACCESS(Pos & c_mypos(), {
	    return c_myinfo().pos() ;
	 }) ;

	 ITK_DEF_ACCESS(Angle & c_mybdir(), {
	    return c_myinfo().bdir() ;
	 }) ;

	 ITK_DEF_ACCESS(Angle & c_myhdir(), {
	    return c_myinfo().hdir() ;
	 }) ;

	 ITK_DEF_ACCESS(Vec & c_myvel(), {
	    return c_myinfo().vel() ;
	 }) ;

	 ITK_DEF_ACCESS(OnlineCoach::SeeInfo::Object & c_ballinfo(), {
	    return coach.sinfo.ball() ;
	 }) ;
	 
	 ITK_DEF_ACCESS(Pos & c_ballpos(), {
	    return c_ballinfo().pos() ;
	 }) ;
	 
	 ITK_DEF_ACCESS(Vec & c_ballvel(), {
	    return c_ballinfo().vel() ;
	 }) ;

	 //--------------------------------------------------
	 // checking by coach's view
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool c_kickable() const {
	    return c_ballpos().distance(c_mypos()) < SC_KICKABLE_AREA ;
	 } ;
   } ;

   //----------------------------------------------------------------------
   // send command and logging

   //--------------------------------------------------
   // turn

   template <class MotherPlayer>
   Bool LogTrajPlayerT<MotherPlayer>::turn(const Flt & power) {
      Bool r = MotherPlayer::turn(power) ;
      lockLog() ; {
	 (*log) << "#PlayerCom["
		<< "step=" << stime().count << ","
		<< "com=turn," 
		<< "power=" << power 
		<< "]" << endl ;
      } ; unlockLog() ;
      return r ;
   }
   
   //--------------------------------------------------
   // dash

   template <class MotherPlayer>
   Bool LogTrajPlayerT<MotherPlayer>::dash(const Flt & power) {
      Bool r = MotherPlayer::dash(power) ;
      lockLog() ; {
	 (*log) << "#PlayerCom["
		<< "step=" << stime().count << ","
		<< "com=dash," 
		<< "power=" << power 
		<< "]" << endl ;
      } ; unlockLog() ;
      return r ;
   }

   //--------------------------------------------------
   // kick

   template <class MotherPlayer>
   Bool LogTrajPlayerT<MotherPlayer>::kick(const Flt & power, 
					   const Flt & dir) {
      Bool r = MotherPlayer::kick(power,dir) ;
      lockLog() ; {
	 (*log) << "#PlayerCom["
		<< "step=" << stime().count << ","
		<< "com=kick," 
		<< "power=" << power << ","
		<< "dir=" << dir
		<< "]" << endl ;
      } ; unlockLog() ;
      return r ;
   }

   //----------------------------------------------------------------------
   // logging coach sensor

   //--------------------------------------------------
   // fork

   template <class MotherPlayer>
   Bool LogTrajPlayerT<MotherPlayer>::forkLogCoachSensor() {
      thrdLogCoachSensor = new Thread(runLogCoachSensor,this) ;
      return True ;
   } ;

   //--------------------------------------------------
   // entry point

   template <class MotherPlayer>
   void * LogTrajPlayerT<MotherPlayer>::runLogCoachSensor(void *p) {
      LogTrajPlayerT<MotherPlayer> * player = 
	 static_cast<LogTrajPlayerT<MotherPlayer> *>(p) ;
      return player->logCoachSensor() ;
   } ;
      
   //--------------------------------------------------
   // main loop

   template <class MotherPlayer>
   void * LogTrajPlayerT<MotherPlayer>::logCoachSensor() {
      CycleStep prev ;
      prev.set(0) ;

      while(True) {
	 coach.cvSee.lock() ; {
	    while(!prev.isOlder(c_time()))
	       coach.cvSee.wait() ;
	    lockLog() ; {
	       // *(log) << *(coach.recvsexp) << endl ; /* for debug */
	       *(log) << "#CoachSee["
		      << "step=" << c_time().count << ","
		      << "mypos=" << c_mypos() << ","
		      << "mybdir=" << c_mybdir() << ","
		      << "myvel=" << c_myvel() << ","
		      << "ballpos=" << c_ballpos() << ","
		      << "ballvel=" << c_ballvel() 
		      << "]" << endl ;
	    } ; unlockLog() ;
	    prev = c_time() ;
	 } ; coach.cvSee.unlock() ;
      }
      return this ;
   } ;

   //----------------------------------------------------------------------
   // logging player sensor

   //--------------------------------------------------
   // fork

   template <class MotherPlayer>
   Bool LogTrajPlayerT<MotherPlayer>::forkLogPlayerSensor() {
      thrdLogPlayerSensor = new Thread(runLogPlayerSensor,this) ;
      return True ;
   } ;

   //--------------------------------------------------
   // entry point

   template <class MotherPlayer>
   void * LogTrajPlayerT<MotherPlayer>::runLogPlayerSensor(void *p) {
      LogTrajPlayerT<MotherPlayer> * player = 
	 static_cast<LogTrajPlayerT<MotherPlayer> *>(p) ;
      return player->logPlayerSensor() ;
   } ;
      
   //--------------------------------------------------
   // main loop

   template <class MotherPlayer>
   void * LogTrajPlayerT<MotherPlayer>::logPlayerSensor() {
      CycleStep prev ;
      prev.set(0) ;

      while(True) {
	 cvSee.lock() ; {
	    while(!prev.isOlder(vtime()))
	       cvSee.wait() ;
	    lockLog() ; {
	       *(log) << "#PlayerSee["
		      << "step=" << vtime().count << ","
		      << "mypos=" << mypos() << ","
		      << "mybdir=" << mybdir() << ","
		      << "myvel=" << myvel() << ","
		      << "ballpos=" << ballpos() << ","
		      << "ballvel=" << ballvel() << ","
		      << "posConf=" << posConf() << ","
		      << "velConf=" << velConf() 
		      << "]" << endl ;
	    } ; unlockLog() ;
	    prev = vtime() ;
	 } ; cvSee.unlock() ;
      }
      return this ;
   } ;

} ;


////////////////////////////////////////////////////////////////////////
#endif	 


	 
	 

	 


