// -*- Mode:C++ -*-
//Header:
//File: Newton2dKFilter.cc
//Title: Bun Player for chagamma
//Author: NODA, Itsuki
//Date: 2002/05/10
//

//ModifyHistory:
// 2002/05/10: Start to create this file
//EndModifyHistory:

#include "Newton2dKFilter.h"

namespace Soccer {

   //======================================================================
   // class ValVector

   //--------------------------------------------------
   ValVector &
   ValVector::
   incMul(const ValMatrix & w, 
	  const ValVector & v,
	  const Bool clearp) {
      
      for(Int i = 0 ; i < size() ; i++) {
	 if(clearp) val(i) = 0.0 ;
	 for(Int j = 0 ; j < size() ; j++) {
	    val(i) += w.val(i,j) * v.val(j) ;
	 }
      } ;
      return *this ;
   } ;

   
   //--------------------------------------------------
   ValVector &
   ValVector::
   decMul(const ValMatrix & w, 
	  const ValVector & v,
	  const Bool clearp) {
      
      for(Int i = 0 ; i < size() ; i++) {
	 if(clearp) val(i) = 0.0 ;
	 for(Int j = 0 ; j < size() ; j++) {
	    val(i) -= w.val(i,j) * v.val(j) ;
	 }
      } ;
      return *this ;
   } ;

   
   //======================================================================
   // class ValVector

   //--------------------------------------------------
   ValMatrix &
   ValMatrix::
   incMul(const ValMatrix & w1,
	  const ValMatrix & w2,
	  const Bool clearp) {

      for(Int i = 0 ; i < sizeN() ; i++) {
	 for(Int j = 0 ; j < sizeM() ; j++) {
	    if(clearp) val(i,j) = 0.0 ;
	    for(Int k = 0 ; k < w1.sizeM() ; k++) {
	       val(i,j) += w1.val(i,k) * w2.val(k,j) ;
	    }
	 }
      } 
      return *this ;
   } ;

   //--------------------------------------------------
   ValMatrix &
   ValMatrix::
   decMul(const ValMatrix & w1,
	  const ValMatrix & w2,
	  const Bool clearp) {

      for(Int i = 0 ; i < sizeN() ; i++) {
	 for(Int j = 0 ; j < sizeM() ; j++) {
	    if(clearp) val(i,j) = 0.0 ;
	    for(Int k = 0 ; k < w1.sizeM() ; k++) {
	       val(i,j) -= w1.val(i,k) * w2.val(k,j) ;
	    }
	 }
      } 
      return *this ;
   } ;

   //======================================================================
   // kalman filter 

   //--------------------------------------------------
   void 
   Newton2dKFilter::
   init() {
      initParam() ;
      initMatrix(True) ;
   } ;

   //--------------------------------------------------
   void 
   Newton2dKFilter::
   initParam() {

      setParam(dfltAlpha, dfltSigmaPos, dfltSigmaVel, dfltSigmaAcc, 
	       dfltSigmaObserv, dfltSize) ;
   } ;

   //--------------------------------------------------
   void 
   Newton2dKFilter::
   setParam(Val a, Val sigmaPos, Val sigmaVel, Val sigmaAcc, 
	    Val sigmaObserv, Int s) {
      alpha() = a ;

      /* <<< [Bun-0.3] <<< */
      setParamPredNoise(sigmaPos,sigmaVel,sigmaAcc,False) ;
      setParamObserv(sigmaObserv,False) ;
      /* >>> [Bun-0.3] >>> */

      size() = s ;
   } ;

   /* <<< [Bun-0.3] <<< */
   //--------------------------------------------------
   void 
   Newton2dKFilter::
   setParamObserv(Val sigmaObserv, Bool setMatrixP) {
      sigmaOPxx() = sigmaObserv ;
      sigmaOPyy() = sigmaObserv ;

      if(setMatrixP) {
	 wR(0,0) = sigmaOPxx() ;
	 wR(1,1) = sigmaOPyy() ;
      } ;

   } ;
   /* >>> [Bun-0.3] >>> */

   /* <<< [Bun-0.3] <<< */
   //--------------------------------------------------
   void 
   Newton2dKFilter::
   setParamPredNoise(Val sigmaPos, Val sigmaVel, Val sigmaAcc, 
		     Bool setMatrixP) {

      sigmaPxx() = sigmaPos ;
      sigmaPyy() = sigmaPos ;
      sigmaVxx() = sigmaVel ;
      sigmaVyy() = sigmaVel ;
      sigmaAxx() = sigmaAcc ;
      sigmaAyy() = sigmaAcc ;

      if(setMatrixP) {
	 wQ(0,0) = sigmaPxx() ; wQ(1,1) = sigmaPyy() ;
	 wQ(2,2) = sigmaVxx() ; wQ(3,3) = sigmaVyy() ;
	 wQ(4,4) = sigmaAyy() ; wQ(5,5) = sigmaAyy() ;
      } ;

   } ;
   /* >>> [Bun-0.3] >>> */

   //--------------------------------------------------
   void 
   Newton2dKFilter::
   allocMatrix() {
      wF().alloc(size(),size()) ;
      wFt().alloc(size(),size()) ;
      wQ().alloc(size(),size()) ;
      wPf().alloc(size(),size()) ;
      wH().alloc(size(),size()) ;
      wR().alloc(size(),size()) ;
      wK().alloc(size(),size()) ;
      wPa().alloc(size(),size()) ;

      wW0().alloc(size(),size()) ;
      wW1().alloc(size(),size()) ;

      vXf().alloc(size()) ;
      vXt().alloc(size()) ;
      vXo().alloc(size()) ;
      vXa().alloc(size()) ;

      vXw().alloc(size()) ;

      /* clear all
       */

      wF().setZero();
      wFt().setZero();
      wQ().setZero();
      wPf().setZero();
      wH().setZero();
      wR().setZero();
      wK().setZero();
      wPa().setZero();

      wW0().setZero();
      wW1().setZero();

      vXf().setZero();
      vXt().setZero();
      vXo().setZero();
      vXa().setZero();

      vXw().setZero();
   } ;

   //--------------------------------------------------
   void 
   Newton2dKFilter::
   initMatrix(Bool allocp) {
      if(allocp) allocMatrix() ;

      /* set wF 
       */
      wF(0,0) = 1.0 ; wF(1,1) = 1.0 ; 
      wF(0,2) = 1.0 ; wF(1,3) = 1.0 ;
      wF(0,4) = 1.0 ; wF(1,5) = 1.0 ;
      wF(2,2) = alpha() ; wF(3,3) = alpha() ; 
      wF(2,4) = alpha() ; wF(3,5) = alpha() ; 

      /* set Ft
       * Ft = F^t
       */
      wFt().setTrans(wF()) ;

      /* set wQ */
      wQ(0,0) = sigmaPxx() ; wQ(1,1) = sigmaPyy() ;
      wQ(2,2) = sigmaVxx() ; wQ(3,3) = sigmaVyy() ;
      wQ(4,4) = sigmaAyy() ; wQ(5,5) = sigmaAyy() ;

      /* set wPf 
       */ 
      // assume zero
      
      /* set wH 
       */
      wH(0,0) = 1.0 ; wH(1,1) = 1.0 ;

      /* set wR 
       */
      wR(0,0) = sigmaOPxx() ;wR(1,1) = sigmaOPyy() ;
   } ;

   //--------------------------------------------------
   void 
   Newton2dKFilter::
   estCycle() {

      /* estimate new Xf
       * Xf = F * Xf
       */
      vXw() = vXf() ;
      vXf().setMul(wF(),vXw()) ;

      /* W0 = F * Pf 
       */
      wW0().setMul(wF(),wPf()) ;

      /* W1 = W0 * F^t 
       */
      wW1().setMul(wW0(),wFt()) ;

      /* Pf = W1 + Q 
       */
      wPf().set(wW1()) ;
      wPf().inc(wQ()) ;
   } ;

   //--------------------------------------------------
   void 
   Newton2dKFilter::
   mergeSensor(Flt px, Flt py) {

      /* set Xo 
       */
      vXo(0) = px ; vXo(1) = py ;

      /* wW0 = (H * Pf * H^t + R)
       * wW1 = wW0^(-1)
       *
       * H $B$O:8>e$N$_C10L9TNs$J$N$G!"(B 
       * H * Pf * H^t $B$O(B Pf $B$r:8>e$@$1%U%#%k%?$7$?$b$N!#(B
       * $BF1$8$/(B R $B$b:8>e$@$1$J$N$G(B,
       * (H * Pf * H^t + R) $B$O:8>e$@$1$7$+@.J,$,$J$$!#(B
       * $B$h$C$F(B (H * Pf * H^t + R)^{-1} $B$N5?;w5U9TNs$O(B
       * $B:8>e$@$1$N@.J,$@$19M$($l$PNI$$!#(B
       */

      wW0().setZero();

      wW0(0,0) = wPf(0,0) + wR(0,0) ; wW0(0,1) = wPf(0,1) + wR(0,1) ;
      wW0(1,0) = wPf(1,0) + wR(1,0) ; wW0(1,1) = wPf(1,1) + wR(1,1) ;

      wW1().setZero() ;

      Flt delta = wW0(0,0) * wW0(1,1) -  wW0(0,1) * wW0(1,0) ;

      wW1(0,0) =   wW0(1,1) / delta ; wW1(0,1) = - wW0(0,1) / delta ; 
      wW1(1,0) = - wW0(1,0) / delta ; wW1(1,1) =   wW0(0,0) / delta ; 

      /* K = Pf * H^t * W1 
       *
       * $B>e$HF1$8$/(B, W1 $B$K$O:8>e$7$+@.J,$,$J$$!#(B
       * $B$J$N$G!"(BPf $B$N:8$H(B, W1 $B$N:8>e$N3]$1;;$r$9$l$PNI$$!#(B
       */

      for(Int i = 0 ; i < size() ; i++) {
	 for(Int j = 0 ; j < 2 ; j++) {
	    wK(i,j) = 0.0 ;
	    for(Int k = 0 ; k < 2 ; k++) {
	       wK(i,j) += wPf(i,k) * wW1(k,j) ;
	    }
	 }
      }

      /* W0 = K * H 
       * W1 = I - W0
       * Pa = W1 * Pf 
       */
      wW0().setMul(wK(),wH()) ;
      wW1().setDiag(1.0) ;
      wW1().dec(wW0()) ;
      wPa().setMul(wW1(),wPf()) ;

      /* Xa = Xf + K * (Xo - H * Xf)
       * $BF1$8$/(B, H $B$O:8>e$N$_(B
       */

      for(Int i = 0 ; i < size() ; i++) {
	 vXa(i) = vXf(i) ;
	 for(Int j = 0 ; j < 2 ; j++) {
	    vXa(i) += wK(i,j) * (vXo(j) - vXf(j)) ;
	 }
      }

      /* Xf = Xa 
       */
      vXf() = vXa() ;

      /* Pf = Pa
       */
      wPf() = wPa() ;
      
   } ;

   //------------------------------------------------------------
   // for test

   //--------------------------------------------------
   inline Flt kfNoise(Flt v) { 
      Flt d = sqrt(3 * v) ;
      return fltRand(-d,d) ;
   } ;

   //--------------------------------------------------
   void 
   Newton2dKFilter::
   testEnvCycle() {

      /* Xt = F * Xt
       */

      vXw() = vXt() ;
      vXt().setMul(wF(),vXw()) ;
	    
      /* Xt += noise
       * [-d:d] $B0lMMJ,I[$NJ,;6$O(B s^2 = (d^2)/3
       * $B$h$C$F!"(Bd = sqrt(3 * s^2)
       */
      vXt(0) += kfNoise(sigmaPxx()) ;
      vXt(1) += kfNoise(sigmaPyy()) ;
      vXt(2) += kfNoise(sigmaVxx() * 0.01) ;
      vXt(3) += kfNoise(sigmaVyy() * 0.01) ;
      
      if(fltRand(1.0) < 0.05) {
	 Flt pow = 10.0 ;
	 vXt(4) = fltRand(-pow,pow) ;
	 vXt(5) = fltRand(-pow,pow) ;
      } else {
	 vXt(4) = 0 ;
	 vXt(5) = 0 ;
      }

   } ;

   //--------------------------------------------------
   Bool
   Newton2dKFilter::
   testCycle() {

      testEnvCycle() ;

      estCycle() ;

      if(fltRand(1.0) < 0.6) {
	 Flt px = vXt(0) + kfNoise(sigmaOPxx()) ;
	 Flt py = vXt(1) + kfNoise(sigmaOPyy()) ;

	 mergeSensor(px,py) ;
	 return True ;
      } else {
	 return False ;
      }

   } ;

   //--------------------------------------------------
   void
   Newton2dKFilter::
   testRun() {
      init() ;
      
      for(Int t = 0 ; t < 10000 ; t++) {
	 Bool r = testCycle() ;

	 cerr << setw(5) << t << "[" << (r ? "*" : " ") << "]:" << endl ;

	 cerr << "vXt()=" ; vXt().showElements(cerr) ; cerr << endl; 
	 cerr << "vXf()=" ; vXf().showElements(cerr) ; cerr << endl; 
	 cerr << "vXo()=" ; vXo().showElements(cerr) ; cerr << endl; 

	 cout << endl ;
      } ;
   } ;
   
} ;

