// -*- Mode:C++ -*-
//Header:
//File: OnlineCoach.cc
//Title: Online Coach for sserver 7
//Author: NODA, Itsuki
//Date: 2001/03/03
//

//ModifyHistory:
// 2001/03/03: Start to create this file
//EndModifyHistory:

#include "OnlineCoach.h"

namespace Soccer {

   //======================================================================
   // Sending init/reconnect

   //--------------------------------------------------
   // send (init...)

   Bool OnlineCoach::sendInit() {
      sendbuf.clear() ;
      sendbuf.printf("(init %s (version %lf))",
		     (char*)teamname,
		     sscproto) ;
      sendbuf.adjustLength() ;

      ITK_VBS(Vbs_Comm,"OnlineCoach:sendInit:send:" << sendbuf ) ;
      send() ;

      recv() ;
      ITK_VBS(Vbs_Comm,"OnlineCoach:sendInit:recv:" << recvbuf) ;

      recvsexp = SimpleSexp::scan(recvbuf,&sxheap) ;
      
      if(isNull(recvsexp)) {
	 ITK_WRN(*this << ": Can't connect with the server.") ;
	 return False ;
      }

      if(!recvsexp->nth(0)->equal("init")) {
	 ITK_WRN(*this << ": Can't initialize the connection." << endl
		 << "server response=" << *recvsexp) ;
	 return False ;
      } 

      if(sscproto < 7) { 
	 if(!recvsexp->nth(1)->equal("ok")) {
	    ITK_WRN(*this << ": Can't initialize the connection." << endl
		    << "server response=" << *recvsexp) ;
	    return False ;
	 } 
      } else {
	 side = scanAsSide(recvsexp->nth(1)) ;
	 if(!recvsexp->nth(2)->equal("ok")) {
	    ITK_WRN(*this << ": Can't initialize the connection." << endl
		    << "server response=" << *recvsexp) ;
	    return False ;
	 } ;
      } 
      
      return True ;

   } ;

   //--------------------------------------------------
   // send (eye...)

   Bool OnlineCoach::sendEye(Bool onoff) {
      Bool r ;
      cvSend.lock() ; {
	 sendbuf.clear() ;
	 sendbuf.printf("(%s %s)",SC_WORDSTR(Eye), 
			(onoff ? SC_WORDSTR(On) : SC_WORDSTR(Off))) ;
	 r = send() ;
	 cvSend.broadcast() ;
      } ; cvSend.unlock() ;
      return r ;
   } ;
		     
   //--------------------------------------------------
   // send (bye)

   Bool OnlineCoach::sendBye() {
      Bool r ;
      cvSend.lock() ; {
	 sendbuf.clear() ;
	 sendbuf.printf("(%s)",SC_WORDSTR(Bye)) ;
	 r = send() ;
	 cvSend.broadcast() ;
      } ; cvSend.unlock() ;
      return r ;
   } ;
		     

   //======================================================================
   // Sensor Loop

   //--------------------------------------------------
   // fork sensor thread

   Thread * OnlineCoach::forkSensor(Bool eyeonp) {
      if(eyeonp) sendEye(True) ;

      thrdSensor = new Thread("onlineCoach.sensor",
			      OnlineCoach::runSensor,this) ;
      return thrdSensor ;
   } 

   //--------------------------------------------------
   // entery for  sensor thread

   void * OnlineCoach::runSensor(void * p) {
      OnlineCoach *coach = static_cast<OnlineCoach*>(p) ;
      return coach->sensorMainLoop() ;
   } ;

   //--------------------------------------------------
   // sensor main loop

   void * OnlineCoach::sensorMainLoop() {
      while(True) {
	 cycleSensor() ;
      } ;
      return this ;
   } ;


   //======================================================================
   // receive 

   //--------------------------------------------------
   // receive from server

   Bool OnlineCoach::cycleSensor() {
      Bool r ;

      r = recv() ;
      ITK_VBS(Vbs_Comm,"OnlineCoach::cycleSensor():recv:" << recvbuf) ;
      if(!r) {
	 ITK_ERR("Fail to receive message from the server.") ;
	 exit(1) ;
	 return False ;
      }
      
      cvSensor.lock() ; {
	 r = scanSensor(SimpleSexp::scan(recvbuf,&sxheap)) ;
	 if(r) cvSensor.broadcast() ;
      } ;cvSensor.unlock() ;
      
      return r ;
   } ;
      
   //--------------------------------------------------
   // receive from server

   Bool OnlineCoach::scanSensor(SimpleSexp & s) {

      UInt len = s.length() ;

      if(len < 1) {
	 ITK_WRN("Illegal Input" << s) ;
	 return False ;
      }

      SimpleSexp * tag = s.nth(0) ;

      Bool r = False ;

      if      (tag->equal(SC_WORDSTR(SeeGlobal))) {
	 cvSee.lock() ; {
	    r = sinfo.scan(s) ;
	    if(r) {
	       cvSee.broadcast() ;
	    }
	 } ; cvSee.unlock() ;
      } 
      else if (tag->equal(SC_WORDSTR(Hear))) {
	 /* ignore */
	 r = True ;
      } 
      else if (tag->equal(SC_WORDSTR(OK))) {
	 /* ignore */
	 r = True ;
      } 
      else if (tag->equal(SC_WORDSTR(ServerParam))) {
	 /* ignore */
	 r = True ;
      } 
      else if (tag->equal(SC_WORDSTR(PlayerParam))) {
	 /* ignore */
	 r = True ;
      } 
      else if (tag->equal(SC_WORDSTR(PlayerType))) {
	 /* ignore */
	 r = True ;
      }
      else {
	 ITK_WRN("Illegal sensor info : sexp=" << s) ;
	 r = False ;
      }
      
      return r ;
   }

   //======================================================================
   // Basic utility

   //--------------------------------------------------
   // send 

   Bool OnlineCoach::send() { 
      socket.send(sendbuf) ;
      sendCount++ ;
      return True ;
   } ;

   //--------------------------------------------------
   // receive

   Bool OnlineCoach::recv() {
      recvbuf.clear() ;
      Bool r = socket.receive(recvbuf) ;
      if(r) recvCount++ ;
      return r ;
   } ;



   //======================================================================
   // See Info

   Bool OnlineCoach::SeeInfo::scan(SimpleSexp & s) {
      UInt len = s.length() ;

      ///// skip "(see" 
      if(len <= 2) return False ;

      ///// time
      time.set(s.nth(1)->scanAsInt()) ;

      ///// object list
      for(SimpleSexp *objlist = s.nthcdr(2) ; objlist->isCons() ; 
	  objlist = objlist->cdr) {
	 SimpleSexp * obj = objlist->car ;
	 SimpleSexp * objname = obj->nth(0) ;
	 SimpleSexp * objtype = objname->nth(0) ;

	 if      (objtype->equal(SC_WORDSTR(Ball)) ||
		  objtype->equal(SC_WORDSTR(SBall))) {
	    ball().pos().set(obj->nth(1)->scanAsFlt(),
			     obj->nth(2)->scanAsFlt()) ;
	    ball().vel().set(obj->nth(3)->scanAsFlt(),
			     obj->nth(4)->scanAsFlt()) ;
	 }
	 else if (objtype->equal(SC_WORDSTR(Player)) ||
		  objtype->equal(SC_WORDSTR(SPlayer))) {
	    Int unum = objname->nth(2)->scanAsInt() ;
	    Object * p ;
	    if(objname->nth(1)->scanAsString() == teamname) {
	       p = &(teammate(unum-1)) ;
	    } else {
	       p = &(opponent(unum-1)) ;
	    }
	    p->pos().set(obj->nth(1)->scanAsFlt(),
			 obj->nth(2)->scanAsFlt()) ;
	    p->vel().set(obj->nth(3)->scanAsFlt(),
			 obj->nth(4)->scanAsFlt()) ;
	    p->bdir() = obj->nth(5)->scanAsFlt() ;
	    p->hdir() = obj->nth(6)->scanAsFlt() ;
	 }
	 else if (objtype->equal(SC_WORDSTR(Goal)) ||
		  objtype->equal(SC_WORDSTR(SGoal))) {
	    /* ignore */
	 }
	 else {
	    ITK_WRN("Unknown visual information:" << *obj) ;
	 }
      }
      return True ;
   } ;

} ;


      
   
