// -*- Mode:C++ -*-
//Header:
//File: OnlineCoach.h
//Title: OnlineCoach interface
//Author: NODA, Itsuki
//Date: 2001/03/12
//

//ModifyHistory:
// 2001/03/12: Start to create this file
//EndModifyHistory:

#ifndef _SPlayer_OnlineCoach_h_
#define _SPlayer_OnlineCoach_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "CycleStep.h"

namespace Soccer {

   using namespace Itk ;

   //----------------------------------------------------------------------
   // constants

   //======================================================================
   // class OnlineCoach

   /*--------------------*/
   /**
    *
    */
   class OnlineCoach : public WithDescriber {
      public:
	 virtual const char * _classname() const { return "OnlineCoach" ; } ;

	 //--------------------------------------------------
	 // constants

	 static const UInt playerN = SC_N_PLAYER_PER_TEAM ;
	 static const UInt playerM = 2 * SC_N_PLAYER_PER_TEAM ;

	 //----------------------------------------
	 // for debug
   
	 static const VbsLevel Vbs_Comm = 5 ;
	 static const VbsLevel Vbs_Test = 1 ;

	 //==================================================
	 // class OnlineCoach::SeeInfo
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 class SeeInfo : public WithDescriber {
	       //========================================
	       // class OnlineCoach::SeeInfo::Object
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       class Object : public WithDescriber {
		     //------------------------------
		     // Type
		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     enum Type {
			Ball,
			Player,
			Other
		     } ;

		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     static const char * typestr(Type type) {
			switch(type) {
			   case Ball:	return "ball" ;
			   case Player: return "player" ;
			   case Other:  return "other" ;
			   default:     return "unknown" ;
			} ;
		     } ;

		     //==============================
		     // body (Object)

		     //------------------------------
		     // member
		  public:
		     Type type ;
		     Int  aux ;
		     PosDir _posdir ;
		     Vec  _vel ;
 
		     //------------------------------
		     // access
		  public:
		     ITK_DEF_ACCESS(PosDir & posdir(), { 
			return _posdir ; 
		     }) ;
		     ITK_DEF_ACCESS(Pos & pos(), { 
			return posdir().pos ; 
		     }) ;
		     ITK_DEF_ACCESS(Angle & bdir(), { 
			return  posdir().bdir ;
		     }) ;
		     ITK_DEF_ACCESS(Angle & hdir(), { 
			return  posdir().hdir ;
		     }) ;
		     ITK_DEF_ACCESS(Vec & vel(), { 
			return  _vel ;
		     }) ;

		     //------------------------------
		     // describe
		  public:
		     void describe(ostream & ostr, 
				   const Bool detailp = True) const {
			ostr << "#Object[" << typestr(type) 
			     << ":" << aux << "]" ;

			if(detailp) {
			   ostr << endl ; 
			   ostr << "\t" << "posdir=" << posdir() << endl ;
			   ostr << "\t" << "vel=" << vel() << endl ;
			}
		     } ;
	       } ;
	       //========================================
	       // body (SeeInfo)

	       //----------------------------------------
	       // member
	    public:
	       Object _ball ;
	       Object _teammate[playerN] ;
	       Object _opponent[playerN] ;
	       
	       SubString teamname ;
	       Side side ;
	       CycleStep time ;
	       
	       SimpleSexp::Heap sxheap ;

	       //----------------------------------------
	       // setup
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       void setup(SubString & tname, Side s) {
		  teamname = tname ;
		  side = s ;

		  ball().type = Object::Ball ;
		  for(UInt i = 0 ; i < playerN ; i++) {
		     teammate(i).type = Object::Player ;
		     teammate(i).aux = i ;
	     
		     opponent(i).type = Object::Player ;
		     opponent(i).aux = i ;
		  }
	       } ;

	       //----------------------------------------
	       // access
	    public:
	       ITK_DEF_ACCESS(Object & ball(), { return _ball ;}) ;
	       
	       ITK_DEF_ACCESS(Object * teammate(), { return _teammate ;});
	       ITK_DEF_ACCESS(Object & teammate(UInt i), {
		  return _teammate[i] ;
	       }) ;

	       ITK_DEF_ACCESS(Object * opponent(), { return _opponent ;});
	       ITK_DEF_ACCESS(Object & opponent(UInt i), {
		  return _teammate[i] ;
	       }) ;

	       //----------------------------------------
	       // scan
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scan(SimpleSexp & s)  ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scan(SimpleSexp * s) { return scan(*s) ; } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scan(Buffer & buffer) {
		  return scan(SimpleSexp::scan(buffer,&sxheap)) ;
	       } ;
	       
	       //----------------------------------------
	       // utility
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isLeft() { return Soccer::isLeft(side) ; } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isRight() { return Soccer::isRight(side) ; } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isMyTeam(SimpleSexp * s) { return s->equal(teamname) ;} ;

	       //----------------------------------------
	       // describe
	    public:
	       void describe(ostream & ostr, 
			     const Bool detailp = True) const {
		  ostr << "#SeeInfo[" << (void *)this <<  "]" ;

		  if(detailp) {
		     ostr << endl ;

		     ostr << "\t" << "ball:" ;
		     ball().describe(ostr,detailp) ;

		     ostr << "\t" << "teammate:" << endl ;
		     for(UInt i = 0 ; i < playerN ; i++) {
			ostr << "\t\t" << "[" << i << "] :" ;
			teammate(i).describe(ostr,detailp) ;
		     }

		     ostr << "\t" << "opponent:" << endl ;
		     for(UInt i = 0 ; i < playerN ; i++) {
			ostr << "\t\t" << "[" << i << "] :" ;
			opponent(i).describe(ostr,detailp) ;
		     }
		  }
	       } ;
	 } ;
   
	 //==================================================
	 // body (OnlineCoach)
   
	 //----------------------------------------
	 // member
      public:
	 SubString teamname ;
	 Side side ;

	 ClientSocket socket ;
	 SubString sschost ;
	 UInt sscport ;
	 Flt sscproto ;
	 
	 Buffer sendbuf ;
	 Buffer recvbuf ;
	 SimpleSexp * recvsexp ;
	 SimpleSexp::Heap sxheap ;
	 
	 CycleStep sendCount ;
	 CycleStep recvCount ;

	 MutexCondVar cvSensor ;
	 MutexCondVar cvSee ;

	 MutexCondVar cvSend ;

	 SeeInfo sinfo ;
	 Thread * thrdSensor ;

	 //----------------------------------------
	 // descructor
      public:
	 virtual ~OnlineCoach() {} ;

	 //----------------------------------------
	 // init
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool init(const char * const shost = DfltSSHost,
		   UInt sport = DfltSSOnCPort,
		   const char * const tname = DfltTeamName,
		   Flt proto = DfltSSProto,
		   Bool openp = True) {
	    sschost = shost ;
	    sscport = sport ;
	    sscproto = proto ;
	    teamname = tname ;
	    side = NoSide ;

	    sendbuf.alloc(DfltBufSize) ;
	    recvbuf.alloc(DfltBufSize) ;

	    Bool r = False ;
	    if(openp) r = open() ;
	    return r ;
	 } ;
	    

	 //----------------------------------------
	 // open connection
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool open() {
	    socket.init((const char *)sschost,Socket::UDP,sscport) ;
	    socket.setBlocking() ;
	    
	    Bool r ;

	    r = sendInit() ;

	    if(r) 
	       sinfo.setup(teamname,side) ;

	    return r ;
	 } ;

	 //----------------------------------------
	 // send 
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool sendInit() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool sendEye(Bool onoff) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool sendBye() ;

	 //----------------------------------------
	 // sensor
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool cycleSensor() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scanSensor(SimpleSexp & s) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scanSensor(SimpleSexp * s) { return scanSensor(*s) ; } ;

	 //----------------------------------------
	 // sensor thread
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Thread * forkSensor(Bool eyeonp = True) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 static void * runSensor(void *p) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void * sensorMainLoop() ;

	 //----------------------------------------
	 // basic utility
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool send() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool recv() ;

	 //----------------------------------------
	 // utility
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool isLeft() { return Soccer::isLeft(side) ; } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool isRight() { return Soccer::isRight(side) ; } ;

	 //----------------------------------------
	 // access
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(CycleStep & time(), { return sinfo.time ; }) ;
	 
	 //----------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#" << _classname() << "[" 
		 << teamname << ":onlineCoach" << "]" ;

	    if(detailp) {
	       ostr << endl ;
	       ostr << "\t" "sshost:port=" 
		    << sschost << ":" << sscport << endl ;
	       ostr << "\t" "protocol_version=" << sscproto << endl ;
	       ostr << "\t" "side=" << sideString(side) << endl ;
	       describeAux(ostr,detailp) ;
	    } 
	 } ;

	 virtual void describeAux(ostream & ostr, 
				  const Bool detailp = True) const {
	    return ;
	 } ;
   } ;


} ;

////////////////////////////////////////////////////////////////////////
#endif

