// -*- Mode:C++ -*-
//Header:
//File: SeeInfo.h
//Title: SeeInfo
//Author: NODA, Itsuki
//Date: 2001/03/05
//

//ModifyHistory:
// 2001/03/05: Start to create this file
//EndModifyHistory:

#ifndef _SPlayer_SeeInfo_h_
#define _SPlayer_SeeInfo_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "CycleStep.h"

namespace Soccer {

   using namespace Itk ;

   //======================================================================
   // SeeInfo class

   /*--------------------*/
   /**
    *
    */
   class SeeInfo : public WithDescriber {

	 //==================================================
	 // Seen Quality
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 enum Quality {
	    SQ_none,
	    SQ_dist,
	    SQ_dir,
	    SQ_dist_chng,
	    SQ_dir_chng,
	    SQ_body_dir,
	    SQ_head_dir
	 } ;

	 //==================================================
	 // Object class
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 class Object : public WithDescriber {

	       //==============================
	       // enum Type
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       enum Type { 
		  Self, 	// see is used in FieldState
		  Ball, 
		  Player, 
		  Goal, 
		  Flag, 
		  Line, 
		  Other } ;

	       //==============================
	       // PlayerID
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       class PlayerID { 

		     //====================
		     // Team enum
		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     enum Team {
			Opponent = -1,
			UnknownTeam = 0,
			OurTeam = 1,
			OtherTeam = 2 // used for misc purpose
		     } ;

		     //====================
		     // const
		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     static const UInt UnknownUNum = 0 ;

		     //====================
		     // body (PlayerID)
		  public:
		     Team team ;
		     UInt unum ;

		     //--------------------
		     // check
		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     Bool isTeammate() const { return team == OurTeam ; } ;

		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     Bool isTeammate(UInt un) const { 
			return isTeammate() && unum == un ; 
		     } ;

		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     Bool isOpponent() const { return team == Opponent ; } ;
		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     Bool isOpponent(UInt un) const { 
			return isOpponent() && unum == un ; 
		     } ;

		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     Bool isTeamUnknown() const { 
			return team == UnknownTeam ; 
		     } ;

		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     Bool isUNumUnknown() const { 
			return unum == UnknownUNum ; 
		     } ;

		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     Bool isUNumKnown() const {
			return unum != UnknownUNum ;
		     } ;

		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     Bool isUnknown() const { 
			return isTeamUnknown() || isUNumUnknown();
		     };
		     
		     //--------------------
		     // describe
		     /*--------------------*/
		     /**
		      *
		      */
		  public:
		     void describe(ostream & ostr, 
					   const Bool detailp = True) const {
			ostr << "#PlayerID[" ;
			if      (isOpponent()) ostr << "-" << unum ;
			else if (isTeammate())  ostr << "+" << unum ;
			else                   ostr << "?" ;
			ostr << "]" ;
			if(detailp) ostr << endl ;
		     } ;
		     
		  public:
		     friend ostream& operator<< (ostream& ostr, 
						 const PlayerID& pid) {
			pid.describe(ostr,False) ;
			return ostr ;
		     } ;
	       } ;

	       //==============================
	       // body (Object)

	       //------------------------------
	       // member
	    public:
	       Type type ;
	       Bool inViewConeP ;
	       union {
		     Landmark* flag ;
		     LineID line ;
		     PlayerID player ;
	       } id ;

	       Quality quality ;

	       Flt dist ;
	       Flt dir ;
	       Flt dist_chng ;
	       Flt dir_chng ;
	       Flt body_dir ;
	       Flt head_dir ;

	       //------------------------------
	       // type check
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isBall() const { return type == Ball ; } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isPlayer() const { return type == Player ; } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isGoal() const { return type == Goal ; } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isFlag() const { return type == Flag ; } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isLine() const { return type == Line ; } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isTeammate() const { 
		  return isPlayer() && id.player.isTeammate() ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isTeammate(UInt unum) const { 
		  return isPlayer() && id.player.isTeammate(unum) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isOpponent() const { 
		  return isPlayer() && id.player.isOpponent() ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isOpponent(UInt unum) const { 
		  return isPlayer() && id.player.isOpponent(unum) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isUnknownPlayer() const {
		  return isPlayer() && id.player.isUnknown() ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isUNumKnown() const {
		  return isPlayer() && id.player.isUNumKnown() ;
	       } ;

	       //------------------------------
	       // short cut
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(UInt team(), { return id.player.team ; }) ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(UInt unum(), { return id.player.unum ; }) ;

	       //------------------------------
	       // seen quality
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool know_dist() const { return quality >= SQ_dist ; }

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool know_dir() const { return quality >= SQ_dir ; }

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool know_dist_chng() const { return quality >= SQ_dist_chng ; }

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool know_dir_chng() const { return quality >= SQ_dir_chng ; }

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool know_body_dir() const { return quality >= SQ_body_dir ; }

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool know_head_dir() const { return quality >= SQ_head_dir ; }

	       //------------------------------
	       // scan
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scan(SimpleSexp & s, 
			 const SubString& teamname) ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scan(SimpleSexp * s, 
			 const SubString& teamname) {
		  return scan(*s,teamname) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scanObjName(SimpleSexp & s, 
				const SubString& teamname) ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scanObjName(SimpleSexp * s, 
				const SubString& teamname) {
		  return scanObjName(*s,teamname) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scan(Buffer& buffer, const SubString& teamname) {
		  return scan(SimpleSexp::scan(buffer), teamname) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scanObjName(Buffer& buffer, const SubString& teamname) {
		  return scanObjName(SimpleSexp::scan(buffer),teamname) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scan(SubString& buffer, const SubString& teamname) {
		  Buffer b = Buffer(buffer) ;
		  return scan(b, teamname) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool scanObjName(SubString& buffer, const SubString& teamname) {
		  Buffer b = Buffer(buffer) ;
		  return scanObjName(b,teamname) ;
	       } ;

	       //------------------------------
	       // relative to absolute position
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool rel2absPos(const PosDir& origin, PosDir& abspos) const ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       PosDir rel2absPos(const PosDir& origin) const {
		  PosDir posdir ;
		  rel2absPos(origin,posdir) ;
		  return posdir ;
	       } ;

	       //------------------------------
	       // relative to absolute velocity
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool rel2absVel(const PosDir& orgpos, 
			       const Vec& orgvel,
			       Vec& absvel) const ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Vec rel2absVel(const PosDir& orgpos, 
			      const Vec& orgvel) const {
		  Vec absvel ;
		  rel2absVel(orgpos,orgvel,absvel) ;
		  return absvel ;
	       } ;

	       //------------------------------
	       // describe
	    public:
	       virtual void describe(ostream & ostr, 
				     const Bool detailp = True) const {
		  ostr << "#Object[" ;
		  if (!inViewConeP) ostr << "?" ;
		  if      (isBall())   { ostr << "ball" ; }
		  else if (isPlayer()) { ostr << id.player ; }
		  else if (isGoal())   { ostr << "goal_" << id.flag->namestr ; }
		  else if (isFlag())   { 
		     ostr << "flag(" << id.flag->namestr << ")" ;
		  } else if (isLine()) { 
		     ostr << "line_" ;
		     switch(id.line) { 
			case LID_l : ostr << "l" ; break ;
			case LID_r : ostr << "r" ; break ;
			case LID_t : ostr << "t" ; break ;
			case LID_b : ostr << "b" ; break ;
			default :       ostr << "?" ; break ;
		     }
		  }
		  ostr << ":dst=" ;
		  if(know_dist()) { ostr << dist ; } else { ostr << "?" ; }
		  if(know_dist_chng()) { ostr << "(" << dist_chng  << ")" ; }

		  ostr << ",dir=" ;
		  if(know_dir()) { ostr << dir ; } else { ostr << "?" ; }
		  if(know_dir_chng()) { ostr << "(" << dir_chng  << ")" ; }

		  if(isPlayer() && know_body_dir() && know_head_dir()) {
		     ostr << ",ang=(B:" << body_dir 
			  << ",H:" << head_dir << ")" ;
		  }
		  ostr << "]" ;
		  if(detailp) ostr << endl ;
	       } ;
	 } ;

	 //==================================================
	 // Main Body (SeeInfo)

	 //--------------------------------------------------
	 // member
      public:
	 CycleStep time ;
	 UInt objN ;
	 Object obj[SC_SEENOBJMAX] ;

	 //--------------------------------------------------
	 // scan
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(SimpleSexp & s, const SubString & teamname) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(SimpleSexp * s, const SubString & teamname) {
	    return scan(*s,teamname) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(Buffer& buffer, const SubString& teamname) {
	    return scan(SimpleSexp::scan(buffer), teamname) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(SubString& buffer, const SubString& teamname) {
	    Buffer b(buffer) ;
	    return scan(b, teamname) ;
	 } ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#SeeInfo[t=" << time << ", n=" << objN << "]" ;
	    if(detailp) {
	       ostr << endl ;
	       for(UInt i = 0 ; i <  objN ; i++) {
		  ostr << "\t" ; 
		  obj[i].describe(ostr,detailp) ;
	       }
	    }
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif
