// -*- Mode:C++ -*-
//Header:
//File: SensorInfo.h
//Title: SeeInfo
//Author: NODA, Itsuki
//Date: 2001/03/05
//

//ModifyHistory:
// 2001/03/05: Start to create this file
//EndModifyHistory:

#ifndef _SPlayer_SensorInfo_h_
#define _SPlayer_SensorInfo_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "CycleStep.h"
#include "SeeInfo.h"
#include "HearInfo.h"
#include "SenseBodyInfo.h"
#include "FullStateInfo.h"

namespace Soccer {

   using namespace Itk ;

   //======================================================================
   // Sensor class

   /*--------------------*/
   /**
    *
    */
   class SensorInfo : public WithDescriber {

	 //==================================================
	 // Type
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 enum Type {
	    T_See = SC_WID_See,
	    T_Hear = SC_WID_Hear,
	    T_SenseBody = SC_WID_SenseBody,
	    T_FullState = SC_WID_FullState,
	    T_ServerParam = SC_WID_ServerParam,
	    T_PlayerParam = SC_WID_PlayerParam,
	    T_PlayerType = SC_WID_PlayerType,
	    T_Error = SC_WID_Error
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 static const char * typeName(Type t) {
	    return ScWordTable[t].name ;
	 } ;
	 
	 //==================================================
	 // Main Body (SensorInfo)

	 //--------------------------------------------------
	 // member

      public:
	 Type type ;
	 CycleStep time ;

	 SeeInfo see ;
	 HearInfo hear ;
	 SenseBodyInfo senseBody ;
	 FullStateInfo fullstate ;
	 
	 SubString teamname ;
	 SubString qteamname ; // quoted teamname

	 //--------------------------------------------------
	 // constructor
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 SensorInfo(char* team = ITK_NULLPTR) { 
	    teamname.set(team) ; 
	 } ;

	 //--------------------------------------------------
	 // set teamname
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void setTeamname(char* team) { 
	    teamname.set(team) ; 
	    Buffer b(DfltNameBufSize) ;
	    b.clear() ;
	    b.printf("\"%s\"",team) ;
	    b.terminate() ;
	    qteamname.copy(b,True) ;
	 } ;

	 //--------------------------------------------------
	 // scan
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(SimpleSexp & s) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(SimpleSexp * s) { return scan(*s) ; } 

	 Bool scan(Buffer & buffer) {
	    return scan(SimpleSexp::scan(buffer)) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(SubString & str) {
	    Buffer b(str) ;
	    return scan(b) ;
	 } ;

	 //--------------------------------------------------
	 // describe
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#SensorInfo[t=" << time 
		 << ",type=" << typeName(type)
		 << "]" ;
	    if(detailp) {
	       ostr << endl ;
	       switch(type) {
		  case T_See:       see.describe(ostr,detailp) ; break ;
		  case T_Hear:      hear.describe(ostr,detailp) ; break ;
		  case T_SenseBody: senseBody.describe(ostr,detailp) ; break ;
		  default : break ;
	       }
	    }
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif
