// -*- Mode:C++ -*-
//Header:
//File: ThreadedPlayer.h
//Title: Threaded Player for Soccer Server 7
//Author: NODA, Itsuki
//Date: 2001/03/07
//

//ModifyHistory:
// 2001/03/07: Start to create this file
//EndModifyHistory:

#ifndef _SPlayer_ThreadedPlayer_h_
#define _SPlayer_ThreadedPlayer_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "BasePlayer.h"
#include "estimatePos.h"
#include "FieldState.h"

namespace Soccer {

   using namespace Itk ;

   //======================================================================
   // Threaded Player

   /*--------------------*/
   /**
    *
    */
   class ThreadedPlayer : public BasePlayer {
      public:
	 virtual const char * _classname() const { return "ThreadedPlayer";};

	 //--------------------------------------------------
	 // member
      public:

	 //------------------------------
	 // sensor thread

	 Thread * thrdSensor ;

	 MutexCondVar cvSensor ;
	 MutexCondVar cvSee ;
	 MutexCondVar cvHear ;
	 MutexCondVar cvSenseBody ;
	 MutexCondVar cvFullState ;

	 //------------------------------
	 // action thread

	 Thread * thrdAction ;

	 MutexCondVar cvSend ;

	 TimeVal prevSendTime ;
	 TimeVal nextSendTime ;

	 TimeVal prevSendBodyTime ;
	 TimeVal nextSendBodyTime ;

	 //--------------------------------------------------
	 // constructor
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ThreadedPlayer() : BasePlayer() {
	    thrdSensor = ITK_NULLPTR ;
	    thrdAction = ITK_NULLPTR ;

	    prevSendTime.update() ;
	    nextSendTime.update() ;

	    prevSendBodyTime.update() ;
	    nextSendBodyTime.update() ;
	 } ;
	 
	 //--------------------------------------------------
	 // destructor
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual ~ThreadedPlayer() {} ;

	 //--------------------------------------------------
	 // send command wrapper
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool sendCommandPre(Bool bodyp)  ;
	 virtual Bool sendCommandPost(Bool bodyp)  ;

	 //--------------------------------------------------
	 // scan sensor wrapper
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool scanSensorPre(SimpleSexp & s) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool scanSensorPost(SimpleSexp & s, Bool result) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool analyzeSensorSee() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool analyzeSensorHear() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool analyzeSensorSenseBody() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool analyzeSensorFullState() ;

	 //--------------------------------------------------
	 // fork sensor thread
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Thread * forkSensor() ;
	    
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 static void * runSensor(void * p) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void * sensorMainLoop() ;
	 
	 //--------------------------------------------------
	 // describe Aux
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void describeAux(ostream & ostr, 
				  const Bool detailp = True) const {
	    return ;
	 } ;
   } ;
} ;

////////////////////////////////////////////////////////////////////////
#endif
